use std::{cell::RefCell, collections::HashMap, fmt::Display, rc::Rc};

use thiserror::Error;

use crate::{
    class::Class,
    token::Token,
    value::{CallableType, Value},
};

#[derive(Error, Debug)]
pub enum InstanceError {
    #[error("line {0}: no property named {1}.")]
    NoSuchProperty(usize, String),
}

#[derive(Debug, Clone, PartialEq)]
pub struct Instance {
    class: Class,
    fields: HashMap<String, Value>,
}

impl Display for Instance {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{} instance", self.class.name)
    }
}

impl Instance {
    pub fn new(class: Class) -> Rc<RefCell<Self>> {
        Rc::new(RefCell::new(Self {
            class,
            fields: HashMap::default(),
        }))
    }

    pub fn get(&self, name: &Token, instance: Rc<RefCell<Self>>) -> Result<Value, InstanceError> {
        if let Some(field) = self.fields.get(name.lexeme.as_str()) {
            Ok(field.clone())
        } else if let Some(method) = self.class.find_method(name.lexeme.as_str()) {
            Ok(Value::Callable((
                Rc::new(method.bind(instance)),
                CallableType::Function,
            )))
        } else {
            Err(InstanceError::NoSuchProperty(
                name.line,
                name.lexeme.clone(),
            ))
        }
    }

    pub fn set(&mut self, name: &Token, value: Value) {
        self.fields.insert(name.lexeme.clone(), value);
    }
}
